;-------------------------- BAC Erat. Sieve, improved ------------------------
;--- J. Arias (2023)

sp=     0xff    ; Stack pointer
lr=     0xfe    ; Link register (return address for leaf subroutines)
tmp16=  0xf0    
number= 0xf2
index=  0xf4
var1=   0xf6
deci0=  0xf7    ; decimal version of 'number'
deci1=  0xf8
deci2=  0xf9
deci3=  0xfa
deci4=  0xfb
; Peripheral registers
UDAT=   0x01    
UFLAGS= 0x02
; Sieve size (up to 0xF0)
;size=   128
;szcmp=  (size-1)>>4

init:   ;lda        0xff         ; init SP
        ;sta        [sp]
        
        lda		4
        ldpg	4
        sta		[tmp16+1]
        lda     0            ; fill sieve with zeroes
        sta     [tmp16]
l01:    decx    [tmp16]
        jnzd    l01
        sta     [x]
        dec		[tmp16+1]
        jnzd	l01
        ldpg	[tmp16+1]
       
        lda     '1'             ; 1 and 2 are primes: print them directly
        sta     [var1]
        lda     .+2
        jmp     cout
        lda     '\n'
        sta     [var1]
        lda     .+2
        jmp     cout
        lda     '2'
        sta     [var1]
        lda     .+2
        jmp     cout
        lda     '\n'
        sta     [var1]
        lda     .+2
        jmp     cout

        lda     0               ; start with number=3
        sta     [number+1]
        sta     [deci3]         ; also init the decimal version of number
        sta     [deci2]
        sta     [deci1]
        lda     3
        sta     [number]
        sta     [deci0]

mbuc:   lda     [number]        ; check bit
        sta     [tmp16]
        lda     [number+1]
        sta     [tmp16+1]
        ror     [tmp16+1]       ; tmp16 >>= 4
        ror     [tmp16]
        lda     [tmp16]         ; var1 = (tmp16 >> 1)
        sta     [var1]
        ror     [tmp16+1]
        ror     [tmp16]
        ror     [tmp16+1]
        ror     [tmp16]
        ror     [tmp16+1]
        ror     [tmp16]
        lda		0xF
        andm	[tmp16+1]
        lda     0x7 
        andm    [var1]          ; var1 = (tmp16 >> 1)&7
        lda     bittab
        addm    [var1]
        jmpd    [var1]          ; program memory table read
        jmpd    .+1
        ldx     [tmp16]
        inc		[tmp16+1]
        ldpg	[tmp16+1]
        tst     [x]             ; bit tested (0 means prime)
        jnz     nxp
        ;------------- number is prime ---------
        ;----- print it -----
        lda     [deci4]         ; Ten-thousands
        jz      pr2
        adda    '0'
        sta     [var1]
        lda     .+2
        jmp     cout
pr2:    lda     [deci3]         ; Thousands
        ora     [deci4]
        jz      pr3
        lda     [deci3]
        adda    '0'
        sta     [var1]
        lda     .+2
        jmp     cout
pr3:    lda     [deci2]         ; Hundreds
        ora     [deci3]
        ora     [deci4]
        jz      pr4
        lda     [deci2]
        adda    '0'
        sta     [var1]
        lda     .+2
        jmp     cout
pr4:    lda     [deci1]         ; Tens
		ora     [deci2]
		ora     [deci3]
		ora     [deci4]
		jz      pr5
		lda     [deci1]
        adda    '0'
        sta     [var1]
        lda     .+2
        jmp     cout
pr5:	lda		[deci0] 		; Units
        adda    '0'
        sta     [var1]
        lda     .+2
        jmp     cout
        
        lda     '\n'            ; separator
        sta     [var1]
        lda     .+2
        jmp     cout
        
;--------------- Now, mark every multiple of number as not prime
        lda     [number]
        sta     [index]
        lda     [number+1]
        sta     [index+1]
l30:    lda     [number]
        addm    [index]
        lda     [number+1]
        adcm    [index+1]
        lda     [index+1]   ; stop if outside table
        cmp     0x3F
        jnc     nxp 
        lda     1           ; skip even indexes
        tst     [index]
        jz      l30

        lda     [index]
        sta     [tmp16]
        lda     [index+1]
        sta     [tmp16+1]
        ror     [tmp16+1]   ; tmp16 >>= 4
        ror     [tmp16]
        lda     [tmp16]
        sta     [var1]      ; var1 = (tmp16 >> 1)
        ror     [tmp16+1]
        ror     [tmp16]
        ror     [tmp16+1]
        ror     [tmp16]
        ror     [tmp16+1]
        ror     [tmp16]
        lda		0xF
        andm	[tmp16+1]    
        lda     0x7
        andm    [var1]      ; var1 = (tmp16 >> 1)&7
        lda     bittab
        addm    [var1]
        jmpd    [var1]      ; program memory table read
        jmpd    .+1
        ldx     [tmp16]
        inc		[tmp16+1]
        ldpg	[tmp16+1]
        orm     [x] 
        jmp     l30
                
nxp:    lda     2           ; next prime= number+2
        addm    [number]
        lda     0
        adcm    [number+1]
        lda     0x40
        cmp     [number+1]
        jpl     stop
        lda     2           ; decimal copy +=2
        addm    [deci0]
        lda     10
        cmp     [deci0]
        jmi     mbuc
        subm    [deci0]
        inc     [deci1]
        cmp     [deci1]
        jmi     mbuc
        subm    [deci1]
        inc     [deci2]
        cmp     [deci2]
        jmi     mbuc
        subm    [deci2]
        inc     [deci3]
        cmp     [deci3]
        jmi     mbuc
        subm	[deci3]
        jmpd	mbuc
        inc		[deci4]
		
stop:   brk                 ; Stop simulation (brk = NOP on real hardware)
        jmp     .

bittab: lda     1           ; table with powers of 2 in program memory
        lda     2
        lda     4
        lda     8
        lda     16
        lda     32
        lda     64
        lda     128

; print character from [var1], "leaf" routine (no stack)
cout:   sta     [lr]
co1:    in      [UFLAGS]        ; wait for TX_rdy (sign bit)
        jpl     co1
        lda     [var1]
        jmpd    [lr]
        out     [UDAT]          ; Transmit data
        jmp     [lr]
        
