`include "bac03.v"
`include "uart_simple.v"

module SYSTEM(
	input clk,
	input reset,
	input rxd,
	output txd,
	output [2:0]led
);

parameter FCLK=96000000/2;

// 1/2 frequency divider with 25% duty cycle
reg ckdiv=0;
always @(negedge clk) ckdiv<=~ckdiv;
wire clk2 = ckdiv & clk;

reg [2:0]outreg;
always @(posedge clk2) if (out&(addr[1:0]==0)) outreg<=outdata[2:0];
assign led=outreg;

wire [7:0]outdata;
wire [7:0]addr;
wire out,in;


bac_computer #(.RAMSIZE(2048), .ROMSIZE(256)) mcore0 ( 
	.clk(clk2), .reset(reset), .din(din), .dout(outdata),
	.out(out), .addr(addr), .in(in)
);
wire [7:0]din = (addr==1) ? uartq : ((addr==2) ? {txrdy,4'bxxxx,rxoverr,rxframeer,rxvalid} : 8'hXX);
wire [7:0]uartq;
wire txrdy,rxvalid,rxoverr,rxframeer;

parameter BAUD = 1500000;
localparam UDIV = (FCLK+BAUD/2)/BAUD;

UART_core #(.DIVIDER(UDIV)) uart0( 
	.clk(clk2), .d(outdata), .wr(out&(addr[1:0]==1)),
	.rd(in&(addr[1:0]==1)), .q(uartq), .txrdy(txrdy),
	.rxvalid(rxvalid), .rxoverr(rxoverr), .rxframeer(rxframeer),
	.nstop(1'b0), .rxd(rxd), .txd(txd)
);

endmodule

